/*
 * Created on 23-Jul-2003
 *
 * 
 */
package jammies.gui;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;

import java.awt.*;
import java.awt.event.*;

import java.io.*;
import java.text.*;

/**
 * @author john
 *
 * 
 */
public class Note extends JWindow {
	private JTextArea textArea = new JTextArea();
	private Color noteColor = new Color( 255, 255, 235 );
	private int borderSize = 3;
	private int dragBarSize = 16;
	private int resizeSize = 10;
	private int colorTabWidth = 12;
	
	private boolean closing = false;
	private boolean dragging = false;
	private boolean resizing  = false;
	private int mouseDX = 0, mouseDY = 0;
	
	private Dimension preferredSize = new Dimension( (colors.length+1)*(colorTabWidth + borderSize) + 32, 128 );
	
	// for inner classes
	private Note self = this; 
	
	private static Frame mainFrame = null;
	
	public Note() {
		super( mainFrame );
		//System.out.println( "making "+getClass().getName() );
		init();
		//System.out.println( "made "+getClass().getName() );
	}
	
	public static void setMainFrame( Frame f ) {
		mainFrame = f;
	}
	
	private void init() {
		getContentPane().add( textArea );
		textArea.setTabSize( 4 );
		textArea.setWrapStyleWord( true );
		textArea.setLineWrap( true );
 
		textArea.getDocument().addDocumentListener(
			new DocumentListener() {
				public void changedUpdate( DocumentEvent de ) {}
			
				public void insertUpdate( DocumentEvent de ) {
					Dimension size = textArea.getSize();
					Dimension prefSize = textArea.getPreferredSize();
					if ( prefSize.height > size.height ) {
						textArea.invalidate();
						setSize( getPreferredSize() );
						validate();
						repaint();
					}
				}

				public void removeUpdate( DocumentEvent de ) {}
			}
		);

		setNoteColor( noteColor );
		//System.out.println( "init");
		setFocusableWindowState( true );
		//setUndecorated( true );
		//enableEvents( Event.MOUSE_UP | Event.GOT_FOCUS | Event.LOST_FOCUS );
		/*editor.addMouseListener( new MouseAdapter() {
			public void mouseReleased( MouseEvent me ) {
				System.out.println( "Requesting Focus" );
				editor.requestFocus();
			}
		});*/

		addMouseListener( new MouseAdapter() {
			public void mousePressed( MouseEvent me ) {	
				if ( !isFocused() ) {
					//textArea.requestFocus();
					return;
				}
				
				int x = me.getX();
				int y = me.getY();
				if ( overResizeCorner( x, y ) ) {
					resizing = true;
				}
				else if ( overCloseBox( x, y ) ) {
					closing = true;
					repaint();
				}
				else {
					dragging = true;
					mouseDX = me.getX();
					mouseDY = me.getY();
				}
		
			}
	
			public void mouseReleased( MouseEvent me ) {
				if ( !textArea.hasFocus() ) {
					textArea.requestFocus();
					return;
				}
								
				if ( overCloseBox( me.getX(), me.getY() ) && closing ) {
					dispatchEvent( new WindowEvent( self, WindowEvent.WINDOW_CLOSING ) );
					return;
				}
				
				if ( overColorTab( me.getX(), me.getY() ) ) {
					Color color = selectedColorTab( me.getX(), me.getY() );
					if ( color != null ) {
						setNoteColor( color );
					}
				}
				
				dragging = false;
				resizing  = false;
				
				if ( closing ) {
					closing = false;
					repaint();
				}
			}
		});

		addMouseMotionListener( new MouseMotionAdapter() {
			public void mouseDragged( MouseEvent me ) {
				if ( dragging ) {
					Point pt = getLocation();
					setLocation( pt.x + me.getX() - mouseDX, pt.y + me.getY() - mouseDY );
				}
				else if ( resizing ) {
					int width  = Math.max( me.getX(), 32 );
					int height = Math.max( me.getY(), 32 );
					setSize( width, height );
					textArea.invalidate();
					validate();
					repaint();
				}
			}

			public void mouseMoved( MouseEvent me ) {
				dragging = false;
				resizing  = false;
				if ( closing ) {
					closing = false;
					repaint();
				}
				//if ( !textArea.hasFocus() )
				//	return;
									
				if ( overResizeCorner( me.getX(), me.getY() ) )
					setCursor( Cursor.getPredefinedCursor( Cursor.SE_RESIZE_CURSOR) );
				else if ( overCloseBox( me.getX(), me.getY() ) || overColorTab( me.getX(), me.getY() ) ) 
					setCursor( Cursor.getPredefinedCursor( Cursor.HAND_CURSOR ) );
				else
					setCursor( Cursor.getDefaultCursor() );
			}
		});


		textArea.addFocusListener( new FocusListener(){
			public void focusGained( FocusEvent fe ) {
				//System.out.println( "Focus gained" );
				setNoteFocused( true );
			}
	
			public void focusLost( FocusEvent fe ) {
				//System.out.println( "Focus lost" );
				setNoteFocused( false );
			}
		});
	}
	
	private boolean isKeyMapped( KeyStroke key, Keymap keyMap ) {
		KeyStroke[] keyStrokes = keyMap.getBoundKeyStrokes();
		
		for ( int j = 0; j < keyStrokes.length; j++ ) {
			KeyStroke mappedKey = keyStrokes[ j ];
			System.out.println( mappedKey );
			if ( mappedKey.getKeyCode() == key.getKeyCode() && mappedKey.getModifiers() == key.getModifiers() )
				return true;
		}
		
		Keymap parent = keyMap.getResolveParent();
		if ( parent != null )
			return isKeyMapped( key, parent );
		return false;
	}
	
	private final static String NOTE_STRING = "Note";
	private final static String COLOR_FORMAT = "Color: {0,number,integer}, {1,number,integer}, {2,number,integer}";
	private final static String LOCATION_FORMAT = "Location: {0,number,integer}, {1,number,integer}";
	private final static String SIZE_FORMAT = "Size: {0,number,integer}, {1,number,integer}";
	
	public void write( OutputStream out ) throws IOException {
		Writer w = new OutputStreamWriter( out );
		w.write( NOTE_STRING+"\n" );
		
		// write color
		Color c = getNoteColor();
		w.write( 
			MessageFormat.format( COLOR_FORMAT, 
				new Object[] { new Integer( c.getRed() ), new Integer( c.getGreen() ), new Integer( c.getBlue() ) } 
			)
			+"\n"
		);
		
		// write location
		Point loc = getLocation();
		w.write( 
			MessageFormat.format( LOCATION_FORMAT, 
				new Object[] { new Integer( loc.x ), new Integer( loc.y ) } 
			)
			+"\n"
		);
		
		// write size
		Dimension size = getSize();
		w.write( 
			MessageFormat.format( SIZE_FORMAT, 
				new Object[] { new Integer( size.width ), new Integer( size.height ) } 
			)
			+"\n"
		);
		
		// write text
		w.write( "\n" );
		String text = getText();
		w.write( text );
		w.flush();
	}
	
	public void read( InputStream in ) throws IOException, ParseException { 
		BufferedReader r = new BufferedReader( new InputStreamReader( in ) );
		String noteLine = r.readLine();
		if ( !noteLine.equals( "Note" ) )
			System.err.println( "Warning: may not be reading a valid note" );
			
		// read color
		String colorLine = r.readLine();
		MessageFormat colorFormat = new MessageFormat( COLOR_FORMAT );
		Object[] components = colorFormat.parse( colorLine );
		int red = ((Number)components[ 0 ]).intValue();
		int green = ((Number)components[ 1 ]).intValue();
		int blue = ((Number)components[ 2 ]).intValue();
		setNoteColor( new Color( red, green, blue ) );
		
		// read location
		String locationLine = r.readLine();
		MessageFormat locationFormat = new MessageFormat( LOCATION_FORMAT );
		Object[] coords = locationFormat.parse( locationLine );
		int x = ((Number)coords[ 0 ]).intValue();
		int y = ((Number)coords[ 1 ]).intValue();
		setLocation( x, y );
		
		//	read size
		String sizeLine = r.readLine();
		MessageFormat sizeFormat = new MessageFormat( SIZE_FORMAT );
		Object[] dims = sizeFormat.parse( sizeLine );
		int width = ((Number)dims[ 0 ]).intValue();
		int height = ((Number)dims[ 1 ]).intValue();
		setSize( width, height );
		
		// read text
		String emptyLine = r.readLine();
		
		StringBuffer text = new StringBuffer();
		String str = null;
		while( (str = r.readLine()) != null ) {
			text.append( str );
			text.append( '\n' );
		}
		setText( text.toString() );
	}
	
	private static final Color[] colors = {
		new Color( 255, 255, 235 ),
		new Color( 235, 255, 255 ),
		new Color( 255, 235, 235 ),
		new Color( 255, 235, 255 ),
		new Color( 235, 255, 235 ),
		new Color( 235, 235, 255 )
	};
	
	
	private boolean overResizeCorner( int x, int y ) {
		int w = getWidth();
		int h = getHeight();
		if ( x <= w && x >= (w-borderSize) ) {
			return y <= h && y >= (h-resizeSize);
		}
		if ( y <= h && y >= (h-borderSize) ) {
			return x <= w && x >= (w-resizeSize);
		}
		return false;
	}
	
	private boolean overCloseBox( int x, int y ) {
		int boxSize = dragBarSize-(2*borderSize);
		return (new Rectangle( borderSize, borderSize, boxSize, boxSize )).contains( x, y );
	}
	
	private boolean overColorTab( int x, int y ) {
		return selectedColorTab( x, y ) != null;
	}
	
	private Color selectedColorTab( int x, int y ) {
		// first see if in drag bar
		if ( y >= borderSize && y <= dragBarSize ) {
			int cx = getWidth() - borderSize - colorTabWidth;
			for ( int i = colors.length-1; i >= 0; i-- ) {
				Color color = colors[ i ];
				if ( new Rectangle( cx, borderSize, colorTabWidth, dragBarSize-borderSize ).contains( x, y ) )
					return color;
				cx -= (colorTabWidth + borderSize);
			}
		}
		return null;
	}
	
	public String getText(){
		return textArea.getText();
	}
	
	public void setText( String text ) {
		textArea.setText( text );
	}
	
	public void setNoteColor( Color color ) {
		this.noteColor = color;
		setBackground( color );
		textArea.setBackground( color );
		repaint();
	}
	
	public Color getNoteColor() {
		return noteColor;
	}
	
	
	public void setSize( int width, int height ) {
		//System.out.println( "setSize("+width+", "+height+")" );
		Dimension size = getSize();
		if ( size.width != width || size.height != height )
			super.setSize( width, height );
		preferredSize = new Dimension( width, height );
	}
	
	public Dimension getPreferredSize() {
		Dimension prefSize = super.getPreferredSize();
		if ( prefSize.width > preferredSize.width )
			preferredSize.width = prefSize.width;
		if ( prefSize.height > preferredSize.height )
			preferredSize.height = prefSize.height;
			
		return preferredSize;
	}
	
	public void setNoteFocused( boolean focused ) {
		//super.setBackground( focused? noteColor.darker() : noteColor );
		repaint();
	}
	
	public Insets getInsets() {
		return new Insets( dragBarSize, borderSize, borderSize, borderSize );
	}
	
	public void paint( Graphics g ) {
		g.setColor( noteColor );
		g.fillRect( 0, 0, getWidth(), getHeight() );
		if ( isFocused() ) {
			g.setColor( noteColor.darker() );
			g.fillRect( 0, 0, getWidth(), dragBarSize );
			g.drawRect( 0, 0, getWidth()-1, getHeight()-1 );
			g.fillRect( getWidth() - resizeSize, getHeight() - resizeSize, resizeSize, resizeSize );
			g.setColor( noteColor );
			int boxSize = dragBarSize-(2*borderSize);
			if ( closing )
				g.fillRect( borderSize, borderSize, boxSize, boxSize );
			else
				g.drawRect( borderSize, borderSize, boxSize, boxSize );

			int x = getWidth() - borderSize - colorTabWidth;
			for ( int i = colors.length-1; i >= 0; i-- ) {
				Color color = colors[ i ];
				g.setColor( color );
				g.fillRect( x, borderSize, colorTabWidth, dragBarSize-borderSize );
				x -= (colorTabWidth + borderSize);
			}

		}
		super.paint( g );
	}
	
	
	/*public static void main(String[] args) {
		mainFrame = new Frame( "Note" );
		Note n = new Note();
		n.addWindowListener( new WindowAdapter() {
			
			public void windowClosing( WindowEvent we ) {
				//System.out.println( "Closing" );
				we.getWindow().dispose();
			}
			
			public void windowClosed( WindowEvent we ) {
				System.exit( 0 );
			}
			
		});
		n.setSize( 200, 200 );
		n.setLocation( 10, 10 );
		n.setNoteColor( new Color( 255, 255, 235 ) );
		n.setText( "Test" );
		mainFrame.pack();
		mainFrame.show();
		n.show();
		
		try {
			java.io.OutputStream out = new java.io.BufferedOutputStream( new java.io.FileOutputStream( "jam-note") );
			java.beans.XMLEncoder encoder = new java.beans.XMLEncoder( out );
			encoder.writeObject( n );
			encoder.flush();
			encoder.close();
		}
		catch( Exception e ) {
			e.printStackTrace();	
		}
	
	}*/
}
